﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using Microsoft.Win32;
using SpotnetWatchdog.Entities;

namespace SpotnetWatchdog.Utils
{
    class Watchdog : IWatchdog
    {
        #region Fields

        private DateTime _previousPostTime;
        private DateTime _lastPostTime;
        private string _spotsFolder;
        private IWatchdogRepository _watchdogRepository;

        private const char PROPERTY_SEPERATOR = '\t';
        private const char TAG_SEPERATOR = '|';

        private const int POST_ID_POS = 0;
        private const int TIMESTAMP_POS = 1;
        private const int CATEGORIE_POS = 3;
        private const int SUBCATEGORIE_POS = 4;
        private const int TITLE_POS = 5;
        private const int SPOTTER_POS = 6;
        private const int TAGS_POS = 7;

        private const string WATCHDOG_REGISTRY_KEY = @"Software\SpotnetWatchdog";
        private const string POST_UPDATE_REGISTRY_KEY = "PreviousUpdateSpots";
        
        #endregion

        #region Constructors

        public Watchdog(IWatchdogRepository watchdogRepository)
        {
            _previousPostTime = GetPreviousPostTime();
            _lastPostTime = _previousPostTime;

            _watchdogRepository = watchdogRepository;
            _spotsFolder = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)
                , "Spotnet");
        }

        #endregion

        #region Public Methods

        public List<WatchResult> GetWatchResults(bool checkNewSpotsOnly)
        {
            Dictionary<int, WatchResult> result = new Dictionary<int, WatchResult>();
            string[] spotFiles = Directory.GetFiles(_spotsFolder, "*.spots");
            DateTime newLastPostTime = DateTime.MinValue;

            //read each file containing spots
            foreach (string spotFile in spotFiles)
            {
                using (StreamReader reader = new StreamReader(spotFile))
                {
                    while (!reader.EndOfStream)
                    {
                        string[] spotProperties = reader.ReadLine().Split(new char[] { PROPERTY_SEPERATOR }, StringSplitOptions.None);
                        
                        if (spotProperties.Length >= 8)
                        {
                            int postId = Int32.Parse(spotProperties[POST_ID_POS]);
                            DateTime postDateTime = DateHelper.GetDateAndTime(Int32.Parse(spotProperties[TIMESTAMP_POS]));

                            //only check if all spots should be checked or if the spot is a new spot
                            if (!checkNewSpotsOnly || postDateTime > _previousPostTime)
                            {
                                //check if post wasn't already added to the results
                                //this is possible when multiple spots files are being checked
                                if (!result.ContainsKey(postId))
                                {
                                    WatchItem matchingWatchItem = GetMatchingWatchItem(spotProperties);
                                    if (matchingWatchItem != null) //a match was found so add spot to result
                                    {
                                        result.Add(postId, CreateWatchResult(spotProperties, matchingWatchItem));
                                    }
                                }
                            }

                            if (postDateTime > newLastPostTime) newLastPostTime = postDateTime;
                        }
                    }
                }
                
            }

            //new spots have been inserted in the spots file so there is a new previous post time we want to store
            if (newLastPostTime > _lastPostTime)
            {
                _previousPostTime = _lastPostTime;
                _lastPostTime = newLastPostTime;

                SavePostTime(_lastPostTime);

                //filter old post if required
                return (from watchResult in result
                        where (!checkNewSpotsOnly || watchResult.Value.DateAndTime > _previousPostTime)
                        //orderby watchResult.Value.Title
                        select watchResult.Value).ToList();
            }

            return result.Values.ToList();
            //return result.Values.OrderBy((wr) => wr.Title).ToList();
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Creates a Watch Result instance
        /// </summary>
        /// <param name="spotProperties"></param>
        /// <param name="watchItem"></param>
        /// <returns></returns>
        private WatchResult CreateWatchResult(string[] spotProperties, WatchItem watchItem)
        {
            //get the category of the spot
            int categoryId = Int32.Parse(spotProperties[CATEGORIE_POS]);
            Category category = _watchdogRepository.Categories.SingleOrDefault((c) => c.Id == categoryId);
            SubCategory subCategory = null;
            List<string> tags = null;

            if (category != null) //a category was found
            {
                //get the sub category of the spot
                int subCategoryId = Int32.Parse(spotProperties[SUBCATEGORIE_POS]);
                subCategory = category.SubCategories.SingleOrDefault((sc) => sc.Id == subCategoryId);

                //get the tags of the spot
                string[] spotTags = spotProperties[TAGS_POS].Split(TAG_SEPERATOR);
                tags = (from tag in category.Tags
                        join spotTag in spotTags on tag.Id equals spotTag
                        where !spotTag.StartsWith("a") //filter sub categories as tags
                        select tag.Name).ToList();
            }

            //create the wacht result instance
            return new WatchResult()
            {
                Category = (category != null ? category.Name : string.Empty),
                DateAndTime = DateHelper.GetDateAndTime(Int32.Parse(spotProperties[TIMESTAMP_POS])),
                Spotter = spotProperties[SPOTTER_POS],
                PostId = Int32.Parse(spotProperties[POST_ID_POS]),
                SubCategory = (subCategory != null ? subCategory.Name : string.Empty),
                Tags = tags,
                Title = spotProperties[TITLE_POS],
                WatchItem = watchItem,
            };
        }

        /// <summary>
        /// Check title and category of spot using watch item
        /// </summary>
        /// <param name="watchItem"></param>
        /// <param name="title"></param>
        /// <param name="titleWords">list with all words in title</param>
        /// <param name="categoryId"></param>
        /// <param name="subCategoryId"></param>
        /// <param name="spotter"></param>
        /// <returns></returns>
        private bool DoesWatchItemMatchSpot(WatchItem watchItem, string title, string[] titleWords, int categoryId, int subCategoryId
            , string spotter)
        {
            //only check title if a search string was specified 
            //and (watch item doesn't filter on category or the category is a match)
            //and (watch item doesn't filter on subcategory or the subcategory is a match)
            //and (watch item doesn't filter on spotter or the spotter is a match)
            if (watchItem.SearchStringWords != null
                && (watchItem.Category == null || watchItem.Category == Category.Empty || watchItem.Category.Id == categoryId)
                && (watchItem.SubCategory == null || watchItem.SubCategory == SubCategory.Empty || watchItem.SubCategory.Id == subCategoryId)
                && (string.IsNullOrWhiteSpace(watchItem.Spotter) || watchItem.Spotter.Equals(spotter, StringComparison.OrdinalIgnoreCase)))
            {
                if (watchItem.WholeWordsOnly)
                {
                    if (watchItem.SearchForAllWords)
                    {
                        return watchItem.SearchStringWords.All((w) => titleWords.Contains(w));
                    }
                    else
                    {
                        return watchItem.SearchStringWords.Any((w) => titleWords.Contains(w));
                    }
                }
                else
                {
                    if (watchItem.SearchForAllWords)
                    {
                        return watchItem.SearchStringWords.All((w) => title.Contains(w));
                    }
                    else
                    {
                        return watchItem.SearchStringWords.Any((w) => title.Contains(w));
                    }
                }
            }

            return false;
        }

        ///<summary>
        /// Retrieves a watch item is the spot matches it's search criteria
        /// </summary>
        /// <param name="spotProperties"></param>
        /// <param name="watchItems"></param>
        /// <returns></returns>
        private WatchItem GetMatchingWatchItem(string[] spotProperties)
        {
            string title = spotProperties[TITLE_POS].ToLower();
            string[] titleWords = title.Split(' ');
            int categoryId = Int32.Parse(spotProperties[CATEGORIE_POS]);
            int subCategoryId = Int32.Parse(spotProperties[SUBCATEGORIE_POS]);
            string spotter = spotProperties[SPOTTER_POS];

            return _watchdogRepository.WatchItems.FirstOrDefault((wi) => 
                DoesWatchItemMatchSpot(wi, title, titleWords, categoryId, subCategoryId, spotter));
        }

        /// <summary>
        /// Retrieves previous update time of spots
        /// </summary>
        /// <returns></returns>
        private DateTime GetPreviousPostTime()
        {
            using (RegistryKey spotnetWatchdogKey = Registry.CurrentUser.OpenSubKey(WATCHDOG_REGISTRY_KEY))
            {
                if (spotnetWatchdogKey != null)
                {
                    return DateTime.Parse(spotnetWatchdogKey.GetValue(POST_UPDATE_REGISTRY_KEY, DateTime.MinValue).ToString()
                        , CultureInfo.InvariantCulture);
                }
            }

            return DateTime.MinValue;
        }

        /// <summary>
        /// Saves new update time spots
        /// </summary>
        /// <param name="updateTimeSpots"></param>
        private void SavePostTime(DateTime updateTimeSpots)
        {
            using (RegistryKey spotnetWatchdogKey = Registry.CurrentUser.CreateSubKey(WATCHDOG_REGISTRY_KEY))
            {
                spotnetWatchdogKey.SetValue(POST_UPDATE_REGISTRY_KEY, updateTimeSpots.ToString(CultureInfo.InvariantCulture));
            }
        }

        #endregion
    }
}
