﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Input;
using GalaSoft.MvvmLight.Command;
using SpotnetWatchdog.Entities;
using SpotnetWatchdog.Utils;
using GalaSoft.MvvmLight.Messaging;
using System.Windows.Data;

namespace SpotnetWatchdog.ViewModels
{
    class WatchResultsViewModel : ViewModelBase
    {
        #region Fields

        private bool _checkNewSpotsOnly;
        private int _numberOfResultsFound;
        private IWatchdog _watchdog;
        private List<WatchResult> _watchResults;
        
        #endregion

        #region Constructors

        public WatchResultsViewModel(bool loadWatchResults)
        {
            _watchdog = new Watchdog(WatchdogRepository.Instance);
            CheckNewSpotsOnly = true;

            InitializeCommands();
            RegisterNotificationMessages();

            if (loadWatchResults) BeginLoadWatchResults(true);
        }

        #endregion

        #region Public Properties

        /// <summary>
        /// Gets or sets CheckNewSpotsOnly
        /// </summary>
        public bool CheckNewSpotsOnly
        {
            get { return _checkNewSpotsOnly; }
            set
            {
                if (_checkNewSpotsOnly != value)
                {
                    _checkNewSpotsOnly = value;
                    RaisePropertyChanged("CheckNewSpotsOnly");
                }
            }
        }

        /// <summary>
        /// Gets or sets NumberOfResultsFound
        /// </summary>
        public int NumberOfResultsFound
        {
            get { return _numberOfResultsFound; }
            set
            {
                if (_numberOfResultsFound != value)
                {
                    _numberOfResultsFound = value;
                    RaisePropertyChanged("NumberOfResultsFound");
                }
            }
        }

        /// <summary>
        /// Gets or sets WatchResults
        /// </summary>
        public List<WatchResult> WatchResults
        {
            get { return _watchResults; }
            private set
            {
                if (_watchResults != value)
                {
                    _watchResults = value;
                    RaisePropertyChanged("WatchResults");
                }
            }
        }
        
        #endregion

        #region Commands

        private void InitializeCommands()
        {
            ReloadWatchResultsCommand = new RelayCommand(() => BeginLoadWatchResults(CheckNewSpotsOnly), () => !IsBusy);
        }

        public ICommand ReloadWatchResultsCommand { get; private set; }
        
        #endregion

        #region Notification Messages

        private void RegisterNotificationMessages()
        {
            Messenger.Default.Register<NewSpotsNotificationMessage>(this,
                (nm) => { BeginLoadWatchResults(CheckNewSpotsOnly); });
        }

        #endregion

        #region Load Watch Results

        private void BeginLoadWatchResults(bool checkNewSpotsOnly)
        {
            if (!IsBusy)
            {
                IsBusy = true;

                BackgroundWorker backgroundWorker = new BackgroundWorker();
                backgroundWorker.DoWork += (sender, e) => e.Result = _watchdog.GetWatchResults((bool)e.Argument);
                backgroundWorker.RunWorkerCompleted += LoadWatchResults_Completed;
                backgroundWorker.RunWorkerAsync(checkNewSpotsOnly);
            }
        }

        private void LoadWatchResults_Completed(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                IsBusy = false;

                MessageDialog.ShowErrorMessage(e.Error.Message);
            }
            else
            {
                ListCollectionView oldWatchResultsView = (ListCollectionView)CollectionViewSource.GetDefaultView(WatchResults);

                WatchResults = (List<WatchResult>)e.Result;
                NumberOfResultsFound = WatchResults.Count;

                if (oldWatchResultsView != null)
                {
                    //add the old sort descriptions to the view of the new watch results collection
                    ListCollectionView watchResultsView = (ListCollectionView)CollectionViewSource.GetDefaultView(WatchResults);
                    foreach (SortDescription sortDesc in oldWatchResultsView.SortDescriptions)
                    {
                        watchResultsView.SortDescriptions.Add(sortDesc);
                    } 
                }

                IsBusy = false;
            }
        }

        #endregion
    }
}